import pickle
import threading
import datetime
import functools

from scooby.plugins_base import Plugin
from scooby.utils import get_stack


threadlocal = threading.local()

def init_threadlocal(threadlocal):
    threadlocal.initialized = True
    threadlocal.memcache_plugin_data = None

init_threadlocal(threadlocal)


class MemcachePluginData(object):
    def __init__(self):
        self.queries = []

    def insert(self, func_name, arg, start, end, time_taken, stack):
        self.queries.append({
            'func_name': func_name,
            'arg': arg,
            'start': start.isoformat(),
            'end': end.isoformat(),
            'time_taken': time_taken,
            'stack': stack,
        })

    def as_json_dict(self):
        return {
            'queries': self.queries,
        }

def record(func):
    @functools.wraps(func)
    def wrapped_method(self, *args, **kwargs):
        if not hasattr(threadlocal, 'initialized'):
            init_threadlocal(threadlocal)

        if not threadlocal.memcache_plugin_data:
            return func(self, *args, **kwargs)

        func_name = func.__name__
        arg = None
        if len(args) > 0:
            arg = args[0]
            if isinstance(arg, dict):
                arg = list(arg.keys())
        start = datetime.datetime.now()
        try:
            return func(self, *args, **kwargs)
        finally:
            end = datetime.datetime.now()
            time_taken = (end - start).total_seconds() * 1000.0
            threadlocal.memcache_plugin_data.insert(
                func_name=func_name,
                arg=arg,
                start=start,
                end=end,
                time_taken=time_taken,
                stack=get_stack()
            )
    return wrapped_method

def getScoobyMemcacheClient():
    import memcache as memc
    class ScoobyMemcacheClient(memc.Client):
        @record
        def flush_all(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).flush_all(*args, **kwargs)

        @record
        def delete_multi(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).delete_multi(*args, **kwargs)

        @record
        def delete(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).delete(*args, **kwargs)

        @record
        def incr(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).incr(*args, **kwargs)

        @record
        def decr(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).decr(*args, **kwargs)

        @record
        def add(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).add(*args, **kwargs)

        @record
        def append(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).append(*args, **kwargs)

        @record
        def prepend(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).prepend(*args, **kwargs)

        @record
        def replace(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).replace(*args, **kwargs)

        @record
        def set(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).set(*args, **kwargs)

        @record
        def cas(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).cas(*args, **kwargs)

        @record
        def set_multi(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).set_multi(*args, **kwargs)

        @record
        def get(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).get(*args, **kwargs)

        @record
        def gets(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).gets(*args, **kwargs)

        @record
        def get_multi(self, *args, **kwargs):
            return super(ScoobyMemcacheClient, self).get_multi(*args, **kwargs)

    return ScoobyMemcacheClient


def patchMemcachedCache(lib):
    from django.core.cache.backends import memcached
    class PatchedLibMemcachedCache(memcached.MemcachedCache):
        def __init__(self, server, params):
            memcached.BaseMemcachedCache.__init__(self, server, params,
                library=lib,
                value_not_found_exception=ValueError)
    memcached.MemcachedCache = PatchedLibMemcachedCache


def patchInitializedCaches(lib):
    from django.core.cache import caches
    cache_classes = caches.all()
    for cache_class in cache_classes:
        cache_class._lib = lib
        if hasattr(cache_class, '_client'):
            client_kwargs = dict(pickleProtocol=pickle.HIGHEST_PROTOCOL)
            client_kwargs.update(cache_class._options)
            cache_class._client = lib.Client(
                cache_class._servers, **client_kwargs)


class MemcachePlugin(Plugin):
    Data = MemcachePluginData

    def __init__(self):
        super(MemcachePlugin, self).__init__(name='Memcache')

    def should_be_used(self):
        try:
            import memcache
            return True
        except ImportError:
            return False

    def instrument(self):
        ScoobyMemcacheClient = getScoobyMemcacheClient()
        class Library(object):
            Client = ScoobyMemcacheClient
        patchMemcachedCache(Library)
        patchInitializedCaches(Library)
        super(MemcachePlugin, self).instrument()

    def on_process_request(self, memcache_plugin_data, *args, **kwargs):
        init_threadlocal(threadlocal)
        threadlocal.memcache_plugin_data = memcache_plugin_data

    def on_process_response(self, *args, **kwargs):
        init_threadlocal(threadlocal)
        threadlocal.memcache_plugin_data = None



import threading
from datetime import datetime

from django.db.models.sql.compiler import SQLCompiler

from scooby.plugins_base import Plugin
from scooby.utils import get_stack


def escaped_string(s):
    return "'%s'" % (
        s.replace('\\', '\\\\').replace("'", "''")
    )

def get_escaped_params(params):
    escaped_params = []
    for param in params:
        if isinstance(param, str):
            escaped_params.append(escaped_string(param))
        elif isinstance(param, bytes):
            escaped_params.append(escaped_string(param.decode("iso-8859-1")))
        elif isinstance(param, bool):
            escaped_params.append(int(param))
        else:
            escaped_params.append(param)
    return tuple(escaped_params)


threadlocal = threading.local()

def init_threadlocal(threadlocal):
    threadlocal.initialized = True
    threadlocal.sql_plugin_data = None

init_threadlocal(threadlocal)

def execute_sql(self, *args, **kwargs):
    if not hasattr(threadlocal, 'initialized'):
        init_threadlocal(threadlocal)

    if not threadlocal.sql_plugin_data:
        return self.execute_sql_default(*args, **kwargs)

    try:
        q, params = self.as_sql()
    except:
        return self.execute_sql_default(*args, **kwargs)

    start = datetime.now()
    try:
        return self.execute_sql_default(*args, **kwargs)
    finally:
        end = datetime.now()
        time_taken = (end - start).total_seconds() * 1000.0
        threadlocal.sql_plugin_data.insert(
            query=q,
            params=params,
            start=start,
            end=end,
            time_taken=time_taken,
            using=self.using,
            stack=get_stack()
        )


class SQLPluginData(object):
    def __init__(self):
        self.queries = []

    def insert(self, query, params, start, end, time_taken, using, stack):
        params = get_escaped_params(params)
        self.queries.append({
            'query_template': query,
            'params': params,
            'query': (query % params),
            'start': start.isoformat(),
            'end': end.isoformat(),
            'time_taken': time_taken,
            'using': using,
            'stack': stack,
        })

    def as_json_dict(self):
        return {
            'queries': self.queries,
        }


class SQLPlugin(Plugin):
    Data = SQLPluginData

    def __init__(self):
        super(SQLPlugin, self).__init__(name='SQL')

    def instrument(self):
        SQLCompiler.execute_sql_default = SQLCompiler.execute_sql
        SQLCompiler.execute_sql = execute_sql
        super(SQLPlugin, self).instrument()

    def on_process_request(self, sql_plugin_data, *args, **kwargs):
        init_threadlocal(threadlocal)
        threadlocal.sql_plugin_data = sql_plugin_data

    def on_process_response(self, *args, **kwargs):
        init_threadlocal(threadlocal)
        threadlocal.sql_plugin_data = None



class Plugin(object):
    def __init__(self, name):
        self.name = name
        self.instrumented = False

    def should_be_used(self):
        return True

    def is_instrumented(self):
        return self.instrumented

    def instrument(self):
        self.instrumented = True



import json

from .plugins_finder import get_plugins


class ScoobyData(object):
    def __init__(self, profiler=None):
        self.profiler = profiler
        self.plugins = get_plugins()
        self.plugins_data = {}
        for plugin in self.plugins:
            self.plugins_data[plugin.name] = plugin.Data()

    def on_process_request(self, request):
        for plugin in self.plugins:
            if hasattr(plugin, 'on_process_request'):
                plugin_data = self.plugins_data[plugin.name]
                plugin.on_process_request(
                    plugin_data, request)

    def on_process_view(self, request, view, view_args, view_kwargs):
        for plugin in self.plugins:
            if hasattr(plugin, 'on_process_view'):
                plugin_data = self.plugins_data[plugin.name]
                plugin.on_process_view(
                    plugin_data, request,
                    view, view_args, view_kwargs)

    def on_process_response(self, request, response):
        for plugin in self.plugins:
            if hasattr(plugin, 'on_process_response'):
                plugin_data = self.plugins_data[plugin.name]
                plugin.on_process_response(
                    plugin_data, request, response)

    def as_json(self):
        plugins_data_json = {}
        for plugin_name in self.plugins_data:
            plugin_data = self.plugins_data[plugin_name]
            plugins_data_json[plugin_name] = plugin_data.as_json_dict()
        return json.dumps({
            'plugins_data': plugins_data_json
        })



from django.apps import AppConfig

from .plugins_finder import get_plugins
from . import settings


class ScoobyAppConfig(AppConfig):
    name = 'scooby'
    verbose_name = 'Scooby'

    def ready(self):
        if not settings.SCOOBY_DEBUG:
            return
        plugins = get_plugins()
        for plugin in plugins:
            if not plugin.is_instrumented():
                plugin.instrument()



import threading

from scooby.plugins_base import Plugin
from scooby.utils import get_stack


threadlocal = threading.local()

def init_threadlocal(threadlocal):
    threadlocal.initialized = True
    threadlocal.log_plugin_data = None

init_threadlocal(threadlocal)


class CustomLogPluginData(object):
    def __init__(self):
        self.logs = []

    def log(self, s, stack):
        self.logs.append({
            's': s,
            'stack': stack,
        })

    def as_json_dict(self):
        return {
            'logs': self.logs,
        }


def log(*args):
    s = ' '.join([(arg if isinstance(arg, str) else repr(arg))
                  for arg in args])
    if not hasattr(threadlocal, 'initialized'):
        return
    if threadlocal.log_plugin_data:
        stack = get_stack()
        threadlocal.log_plugin_data.log(s, stack)


class CustomLogPlugin(Plugin):
    Data = CustomLogPluginData

    def __init__(self):
        super(CustomLogPlugin, self).__init__(name='CustomLog')

    def on_process_request(self, log_plugin_data, *args, **kwargs):
        init_threadlocal(threadlocal)
        threadlocal.log_plugin_data = log_plugin_data

    def on_process_response(self, *args, **kwargs):
        init_threadlocal(threadlocal)
        threadlocal.log_plugin_data = None



import threading
from datetime import datetime

from scooby.plugins_base import Plugin
from scooby.utils import get_stack


threadlocal = threading.local()

def init_threadlocal(threadlocal):
    threadlocal.initialized = True
    threadlocal.thriftpy_plugin_data = None

init_threadlocal(threadlocal)


def parse_args(args):
    parsed = []
    for v in args:
        parsed.append(repr(v))
    return parsed

def parse_kwargs(kwargs):
    parsed = {}
    for k, v in kwargs.items():
        if isinstance(v, bytes):
            v = str(v)
        parsed[k] = v
    return parsed

orig_req = None
def req(self, api, *args, **kwargs):
    if not hasattr(threadlocal, 'initialized'):
        init_threadlocal(threadlocal)

    if not threadlocal.thriftpy_plugin_data:
        return orig_req(self, api, *args, **kwargs)

    start = datetime.now()
    try:
        return orig_req(self, api, *args, **kwargs)
    finally:
        end = datetime.now()
        time_taken = (end - start).total_seconds() * 1000.0
        threadlocal.thriftpy_plugin_data.insert(
            service_name=self._service.__name__,
            func_name=api,
            args=args,
            kwargs=kwargs,
            start=start,
            end=end,
            time_taken=time_taken,
            stack=get_stack()
        )


class ThriftpyPluginData(object):
    def __init__(self):
        self.queries = []

    def insert(self, service_name, func_name, args, kwargs, start, end, time_taken, stack):
        self.queries.append({
            'service_name': service_name,
            'func_name': func_name,
            'args': parse_args(args),
            'kwargs': parse_kwargs(kwargs),
            'start': start.isoformat(),
            'end': end.isoformat(),
            'time_taken': time_taken,
            'stack': stack,
        })

    def as_json_dict(self):
        return {
            'queries': self.queries,
        }


class ThriftpyPlugin(Plugin):
    Data = ThriftpyPluginData

    def __init__(self):
        super(ThriftpyPlugin, self).__init__(name='Thriftpy')

    def should_be_used(self):
        try:
            import thriftpy
            return True
        except ImportError:
            return False

    def instrument(self):
        from thriftpy.thrift import TClient
        global orig_req
        orig_req = TClient._req
        TClient._req = req
        super(ThriftpyPlugin, self).instrument()

    def on_process_request(self, thriftpy_plugin_data, *args, **kwargs):
        init_threadlocal(threadlocal)
        threadlocal.thriftpy_plugin_data = thriftpy_plugin_data

    def on_process_response(self, *args, **kwargs):
        init_threadlocal(threadlocal)
        threadlocal.thriftpy_plugin_data = None



from scooby.plugins_base import Plugin


class ViewNamePluginData(object):
    def __init__(self):
        self.view_name = None
        self.args = ()
        self.kwargs = {}

    def as_json_dict(self):
        return {
            'view_name': self.view_name,
            'args': self.args,
            'kwargs': self.kwargs,
        }


class ViewNamePlugin(Plugin):
    Data = ViewNamePluginData

    def __init__(self):
        super(ViewNamePlugin, self).__init__(name='ViewName')

    def on_process_view(self, plugin_data, request,
            view, view_args, view_kwargs):
        plugin_data.view_name = '%s.%s' % (view.__module__, view.__name__)
        plugin_data.args = view_args
        plugin_data.kwargs = view_kwargs



import os
import jwt
import uuid
import cProfile
from datetime import datetime
from django.utils.deprecation import MiddlewareMixin

from .base import ScoobyData
from .utils import get_backend

from . import settings


class ScoobyMiddleware(MiddlewareMixin):
    def get_cookie_config(self, request):
        if not settings.SCOOBY_DEBUG:
            return
        encoded_jwt = request.COOKIES.get(settings.SCOOBY_COOKIE_NAME)
        if encoded_jwt is None:
            return
        try:
            return jwt.decode(encoded_jwt, settings.SCOOBY_SECRET_KEY)
        except:
            pass

    def process_request(self, request):
        cookie_config = self.get_cookie_config(request)
        if cookie_config is None or not cookie_config.get('enable'):
            return
        profiler = None
        if cookie_config.get('cprofile'):
            profiler = cProfile.Profile()
            profiler.enable()
        request.scooby_data = ScoobyData(profiler)
        request.scooby_data.on_process_request(request)

    def process_view(self, request, view, view_args, view_kwargs):
        cookie_config = self.get_cookie_config(request)
        if cookie_config is None or not cookie_config.get('enable'):
            return
        if hasattr(request, 'scooby_data'):
            request.scooby_data.on_process_view(
                request, view, view_args, view_kwargs)

    def process_response(self, request, response):
        cookie_config = self.get_cookie_config(request)
        if cookie_config is None or not cookie_config.get('enable'):
            return response
        if not hasattr(request, 'scooby_data'):
            return response
        request.scooby_data.on_process_response(request, response)
        unique_hex = uuid.uuid4().hex
        response['X-Scooby'] = unique_hex
        start = datetime.now()
        # Set data in backend for 2 mins.
        backend = get_backend()
        backend.set(unique_hex, request.scooby_data.as_json(), 120)
        if cookie_config.get('cprofile'):
            # Set cProfile stats in backend for 2 mins.
            profiler = request.scooby_data.profiler
            profiler.disable()
            temp_stats_filename = '/etc/%s.stats' % unique_hex
            profiler.dump_stats(temp_stats_filename)
            f = open(temp_stats_filename, 'rb')
            stats_key = unique_hex + '-pstats'
            backend.set(stats_key, f.read(), 120)
            f.close()
            os.remove(temp_stats_filename)
        end = datetime.now()
        time_taken = round((end - start).total_seconds() * 1000.0)
        response['X-Scooby-overhead'] = time_taken
        return response



from . import settings


class RedisConfigNotProvided(Exception):
    pass

class RedisBackend(object):
    def __init__(self):
        import redis
        if settings.SCOOBY_REDIS_BACKEND_CONFIG is None:
            raise RedisConfigNotProvided(
                'SCOOBY_REDIS_BACKEND_CONFIG is not set in settings')
        self.redis = redis.StrictRedis(**settings.SCOOBY_REDIS_BACKEND_CONFIG)

    def get(self, key):
        return self.redis.get(key)

    def set(self, key, value, timeout=None):
        return self.redis.set(key, value, timeout)

    def delete(self, key):
        return self.redis.delete(key)



from django.http import HttpResponse

from .utils import get_backend


def get_data(request, uuid):
    backend = get_backend()
    json_data = backend.get(uuid) or ''
    backend.delete(uuid)
    return HttpResponse(json_data, content_type="application/json")


def get_cprofile_data(request, uuid):
    backend = get_backend()
    stats = backend.get(uuid + '-pstats')
    if stats is None:
        return HttpResponse('No stats found.')
    filename = request.GET['filename']
    response = HttpResponse(stats)
    response['Content-Disposition'] = 'attachment; filename=%s' % filename
    return response



from .plugins.viewname import ViewNamePlugin
from .plugins.sql import SQLPlugin
from .plugins.processtime import ProcessTimePlugin
from .plugins.customlog import CustomLogPlugin
from .plugins.memcache import MemcachePlugin
from .plugins.thriftpy import ThriftpyPlugin

plugin_instances = [
    ViewNamePlugin(),
    ProcessTimePlugin(),
    SQLPlugin(),
    CustomLogPlugin(),
    MemcachePlugin(),
    ThriftpyPlugin(),
]

plugins_to_be_used = [p for p in plugin_instances if p.should_be_used()]

def get_plugins():
    return plugins_to_be_used



from django.conf import settings


SCOOBY_DEBUG = getattr(settings, 'SCOOBY_DEBUG', settings.DEBUG)
SCOOBY_BACKEND = getattr(settings, 'SCOOBY_BACKEND',
                         'scooby.backends.RedisBackend')
SCOOBY_REDIS_BACKEND_CONFIG = getattr(settings, 'SCOOBY_REDIS_BACKEND_CONFIG',
                                      None)
SCOOBY_SECRET_KEY = settings.SCOOBY_SECRET_KEY
SCOOBY_COOKIE_NAME = 'scoobydoobydoo'



import datetime

from scooby.plugins_base import Plugin


class ProcessTimePluginData(object):
    def __init__(self):
        self.start = None
        self.end = None

    def as_json_dict(self):
        process_time = None
        if self.start and self.end:
            process_time = int((self.end - self.start).total_seconds() * 1000)
        return {
            'start': self.start and self.start.isoformat(),
            'end': self.end and self.end.isoformat(),
            'process_time': process_time,
        }


class ProcessTimePlugin(Plugin):
    Data = ProcessTimePluginData

    def __init__(self):
        super(ProcessTimePlugin, self).__init__(name='ProcessTime')

    def on_process_request(self, plugin_data, *args, **kwargs):
        plugin_data.start = datetime.datetime.now()

    def on_process_response(self, plugin_data, *args, **kwargs):
        plugin_data.end = datetime.datetime.now()



from setuptools import setup, find_packages

# read the contents of README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md')) as f:
        long_description = f.read()

# setup the project
setup(
    name="django-scooby-profiler",
    version="2.0.4",
    url="https://github.com/shhaumb/django-scooby-profiler",
    author="Shubham Jain",
    author_email="sj.iitr@gmail.com",
    description=(
        "A debugging tool for Django applications which works for all HTTP "
        "requests including AJAX. Using this, you can profile Django views. "
        "It shows you all SQL, Memcache queries with proper stacktrace "
        "happening in app while serving a request, with the help of "
        "'Scooby profiler' chrome extension."
    ),
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords="django debugging profiler profiling sql tool",
    license="MIT",
    packages=find_packages(),
    install_requires=[
        'redis',
        'pyjwt',
    ]
)



from django.conf.urls import url

from .views import get_data, get_cprofile_data


urlpatterns = [
    url(r'^get-data/(?P<uuid>[\w\-]+)/$', get_data, name='scooby_get_data'),
    url(r'^get-cprofile-data/(?P<uuid>[\w\-]+)/$', get_cprofile_data,
        name='scooby_get_cprofile_data'),
]



import os
import sys
import inspect
import importlib

from django.db.models import Model

from . import settings


global_backend = None
def get_backend():
    global global_backend
    if global_backend is None:
        backend_module_path, backend_class_name = settings.SCOOBY_BACKEND.rsplit('.', 1)
        backend_module = importlib.import_module(backend_module_path)
        backend_class = getattr(backend_module, backend_class_name)
        global_backend = backend_class()
    return global_backend

def curate_filename(filename):
    for path in reversed(sys.path):
        if ('dist-packages' in path or 'site-packages' in path or
            '/usr/lib/python' in path):
            continue
        if path and filename.startswith(path):
            filename = filename.split(path, 1)[1]
            if filename.startswith(os.path.sep):
                filename = filename[1:]
            break
    return filename


def get_repr(value):
    try:
        unicode_cls = unicode
    except NameError:
        # Python3
        unicode_cls = str
    if isinstance(value, (
            int, float, bool, str, unicode_cls, type(None))):
        return repr(value)
    elif isinstance(value, Model):
        has_title_field = False
        try:
            value._meta.get_field('title')
            has_title_field = True
        except:
            pass
        if has_title_field:
            return "<%s: pk=%s title=%s>" % (
                value.__class__.__name__,
                get_repr(value.pk),
                get_repr(value.title),
            )
        return "<%s: pk=%s>" % (
            value.__class__.__name__,
            get_repr(value.pk),
        )
    elif isinstance(value, tuple):
        return "(%s)" % (', '.join([get_repr(item) for item in value]))
    elif isinstance(value, list):
        return "[%s]" % (', '.join([get_repr(item) for item in value]))
    elif isinstance(value, dict):
        return "{%s}" % (', '.join([
            '%s: %s' % (get_repr(key), get_repr(value)) for (key, value) in
            value.items()]))
    else:
        try:
            return "<%s>" % (value.__class__.__name__)
        except:
            return "<<Some object>>"

def get_stack():
    frame_infos = inspect.stack(9)
    stack = []
    for frame_info in frame_infos[2:]:
        frame, filename, lineno, function, code_context, index = frame_info
        frame_locals = inspect.getargvalues(frame).locals
        local_vars = {}
        if not ('dist-packages' in filename or
                'site-packages' in filename or
                '/usr/lib/python' in filename):
            for var in frame_locals:
                value = get_repr(frame_locals[var])
                local_vars[var] = value
        stack.append({
            'filename': curate_filename(filename),
            'lineno': lineno,
            'function': function,
            'code_context': code_context,
            'line_index': index,
            'local_vars': local_vars,
        })
        del frame_locals
        del frame
    del frame_infos
    return stack
